// ----------------------------------------------------------------------------
//
// rt_tbag.i
// RT++ thread bags
//
// $Id: rt_tbag.i,v 1.9 1996/03/26 12:36:31 schreine Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------
#ifndef RT_TBAG_I
#define RT_TBAG_I

// ----------------------------------------------------------------------------
//
// ThreadBagBase<T>: a bag of threads with results of type <T>
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//
// initialize empty bag
//
// ----------------------------------------------------------------------------
template <class T>
inline void 
ThreadBagBase<T>::init(void)
{
  ThreadBagCore::init();
  result_list = 0;
  result_tail = 0;
  result_array = 0;
}

// ----------------------------------------------------------------------------
//
// close bag with <k> more threads to be added
//
// ----------------------------------------------------------------------------
template <class T>
inline void 
ThreadBagBase<T>::close(int k)
{
  const int n = threads+k;
  result_array = (T*)processor.alloc(n*sizeof(T));
  ThreadBagCore::close(k);
  result_tail = 0;
  for (int i = 0; i < n; i++)
    {
      if (result_list)
	{
	  result_array[i] = result_list->head;
	  result_list = result_list->tail;
	  continue;
	}
      void (*f)(T*) = Type<T>::init;
      if (!f) break;
      for (int j = i; j < n; j++)
	f(result_array+j);
      break;
    }
  mutex.unlock();
  processor.unlock();
}

// ----------------------------------------------------------------------------
//
// close bag (no more threads are added)
//
// ----------------------------------------------------------------------------
template <class T>
inline void 
ThreadBagBase<T>::close(void)
{
  close(0);
}

// ----------------------------------------------------------------------------
//
// return result of thread number <n> and set <d> to 1, if such a result exists
//
// ----------------------------------------------------------------------------
template <class T>
inline void
ThreadBagBase<T>::wait(const int n, int *d, T *r)
{
  while (n >= results) 
    {
      if (!closed || n < threads) 
	{
	  ThreadBagCore::wait(n);
	}
      if (n >= threads)
	{
	  *d = 0;
	  return;
	}
    }
  *d = 1;
  if (closed) 
    { 
      *r = result_array[n];
      return;
    }
  processor.lock();
  mutex.lock();
  if (closed)
    {
      processor.unlock();
      mutex.unlock();
      *r = result_array[n];
      return;
    }
  ResultList<T> *list = result_list;
  for (int i = 0; i < n; i++)
    list = list->tail;
  *r = list->head;
  processor.unlock();
  mutex.unlock();
}

// ----------------------------------------------------------------------------
//
// deliver result
//
// ----------------------------------------------------------------------------
template <class T>
inline void 
ThreadBagBase<T>::deliver(CoreBagThread *t, T r)
{
  ResultList<T> *cell = 0;
  if (!closed)
    {
      cell = (ResultList<T>*)processor.alloc(sizeof(ResultList<T>));
      cell->head = r;
      cell->tail = 0;
    }
  int a = ThreadBagCore::prepare();
  if (closed)
    result_array[a] = r;
  else
    {
      if (result_tail)
	result_tail->tail = cell;
      else
	result_list = cell;
      result_tail = cell;
    }
  ThreadBagCore::deliver(t);
}

// ----------------------------------------------------------------------------
//
// scan function
//
// ----------------------------------------------------------------------------
template <class T>
inline void 
ThreadBagBase<T>::scan(void)
{
  ThreadBagCore::scan();                   // scan core thread
  void(*const f)(T*) = Type<T>::scan;      // scan result_list
  if (f)
    {
      ResultList<T> *p = result_list;
      while (p)
	{
	  if (Block::block(p)->mark()) return;
	  f(&p->head);
	  p = p->tail;
	}
    }
  else
    {
      ResultList<T> *p = result_list;
      while (p)
	{
	  if (Block::block(p)->mark()) return;
	  p = p->tail;
	}
    }
  T *p = result_array;                     // scan result_array
  if (!p) return;           
  if (Block::block(p)->mark()) return;
  void (*const g)(T *) = Type<T>::scan; 
  if (!g) return;
  T *l = p+threads;
  while (p < l)
    {
      g(p);
      p++;        
    }
}

// ----------------------------------------------------------------------------
//
// ThreadBag<R>: bag pointer typed in result only
//
// ----------------------------------------------------------------------------

template <class R>
inline
ThreadBag<R>::ThreadBag(void): Pointer< ThreadBagBase<R> >()
{
}

template <class R>
inline
ThreadBag<R>::ThreadBag(int n): Pointer< ThreadBagBase<R> >(1)
{
  pointer->close(n);
}

template <class R>
inline void 
ThreadBag<R>::open(void)
{
  alloc();
}

template <class R>
inline void 
ThreadBag<R>::open(int n)
{
  alloc();
  pointer->close(n);
}

template <class R>
inline ThreadBag<R>
ThreadBag<R>::Open(void)
{
  ThreadBag bag;
  bag.open();
  return(bag);
}

template <class R>
inline ThreadBag<R>
ThreadBag<R>::Open(int n)
{
  ThreadBag bag;
  bag.open(n);
  return(bag);
}

template <class R>
inline R 
Wait(ThreadBag<R> bag, int i)
{
  int done;
  R result;
  bag->wait(i, &done, &result);
  if (!done) rt_abort("thread bag index too large");
  return(result);
}

template <class R>
inline R 
Wait(ThreadBag<R> bag, int i, int *done)
{
  R result;
  bag->wait(i, done, &result);
  return(result);
}

template <class R>
inline R 
Size(ThreadBag<R> bag)
{
  return(bag->size());
}

// ----------------------------------------------------------------------------
//
// ResBagThread<R>: a bag thread with result of type <R>
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//
// add thread to bag
//
// ----------------------------------------------------------------------------
template <class R> 
inline void 
ResBagThread<R>::add(ThreadBag<R> _bag)
{
  _bag->add(this);     // add thread to bag
  bag = _bag;          // our result bag
}

// ----------------------------------------------------------------------------
//
// scan function for bag threads
//
// ----------------------------------------------------------------------------
template <class R> 
inline void 
ResBagThread<R>::scan(void)
{
  CoreBagThread::scan();
  bag.scan();
}

// ----------------------------------------------------------------------------
//
// ArgBagThread0<R>: bag thread with zero arguments
//
// ----------------------------------------------------------------------------

template <class R>
inline void 
ArgBagThread0<R>::user(void)
{
  bag->deliver(this, f());
}

template <class R>
inline void 
ArgBagThread0<R>::scan(void)
{
  ResBagThread<R>::scan();
}

template <class R>
inline void 
ArgBagThread0<R>::init(ThreadBag<R> _bag, R (*_f)(void))
{
  ResBagThread<R>::init();
  f = _f;
  userfun = (void(*)(CoreThread*))user;
  scanfun = (void(*)(CoreThread*))ArgBagThread0<R>::scan;
  add(_bag);
}

// ----------------------------------------------------------------------------
//
// ArgBagThread1<R, A1>: bag thread with one argument
//
// ----------------------------------------------------------------------------

template <class R, class A1>
inline void 
ArgBagThread1<R, A1>::user(void)
{
  bag->deliver(this, f(a1));
}

template <class R, class A1>
inline void 
ArgBagThread1<R, A1>::scan(void)
{
  ResBagThread<R>::scan();
  void(*const f1)(A1*) = Type<A1>::scan;
  if (f1) f1(&a1);
}

template <class R, class A1>
inline void 
ArgBagThread1<R, A1>::init(ThreadBag<R> _bag, R (*_f)(A1), A1 _a1)
{
  ResBagThread<R>::init();
  f = _f;
  a1 = _a1;
  userfun = (void(*)(CoreThread*))user;
  scanfun = (void(*)(CoreThread*))ArgBagThread1<R, A1>::scan;
  add(_bag);
}

// ----------------------------------------------------------------------------
//
// ArgBagThread2<R, A1, A2>: bag thread with two arguments
//
// ----------------------------------------------------------------------------

template <class R, class A1, class A2>
inline void 
ArgBagThread2<R, A1, A2>::user(void)
{
  bag->deliver(this, f(a1, a2));
}

template <class R, class A1, class A2>
inline void 
ArgBagThread2<R, A1, A2>::scan(void)
{
  ResBagThread<R>::scan();
  void(*const f1)(A1*) = Type<A1>::scan;
  if (f1) f1(&a1);
  void(*const f2)(A2*) = Type<A2>::scan;
  if (f2) f2(&a2);
}

template <class R, class A1, class A2>
inline void 
ArgBagThread2<R, A1, A2>::
init(ThreadBag<R> _bag, R (*_f)(A1, A2), A1 _a1, A2 _a2)
{
  ResBagThread<R>::init();
  f = _f;
  a1 = _a1;
  a2 = _a2;
  userfun = (void(*)(CoreThread*))user;
  scanfun = (void(*)(CoreThread*))ArgBagThread2<R, A1, A2>::scan;
  add(_bag);
}

// ----------------------------------------------------------------------------
//
// ArgBagThread3<R, A1, A2, A3>: bag thread with three arguments
//
// ----------------------------------------------------------------------------

template <class R, class A1, class A2, class A3>
inline void 
ArgBagThread3<R, A1, A2, A3>::user(void)
{
  bag->deliver(this, f(a1, a2, a3));
}

template <class R, class A1, class A2, class A3>
inline void 
ArgBagThread3<R, A1, A2, A3>::scan(void)
{
  ResBagThread<R>::scan();
  void(*const f1)(A1*) = Type<A1>::scan;
  if (f1) f1(&a1);
  void(*const f2)(A2*) = Type<A2>::scan;
  if (f2) f2(&a2);
  void(*const f3)(A2*) = Type<A3>::scan;
  if (f3) f3(&a3);
}

template <class R, class A1, class A2, class A3>
inline void 
ArgBagThread3<R, A1, A2, A3>::
init(ThreadBag<R> _bag, R (*_f)(A1, A2, A3), A1 _a1, A2 _a2, A3 _a3)
{
  ResBagThread<R>::init();
  f = _f;
  a1 = _a1;
  a2 = _a2;
  a3 = _a3;
  userfun = (void(*)(CoreThread*))user;
  scanfun = (void(*)(CoreThread*))ArgBagThread3<R, A1, A2, A3>::scan;
  add(_bag);
}

// ----------------------------------------------------------------------------
//
// ArgBagThread4<R, A1, A2, A3, A4>: bag thread with four arguments
//
// ----------------------------------------------------------------------------

template <class R, class A1, class A2, class A3, class A4>
inline void 
ArgBagThread4<R, A1, A2, A3, A4>::user(void)
{
  bag->deliver(this, f(a1, a2, a3, a4));
}

template <class R, class A1, class A2, class A3, class A4>
inline void 
ArgBagThread4<R, A1, A2, A3, A4>::scan(void)
{
  ResBagThread<R>::scan();
  void(*const f1)(A1*) = Type<A1>::scan;
  if (f1) f1(&a1);
  void(*const f2)(A2*) = Type<A2>::scan;
  if (f2) f2(&a2);
  void(*const f3)(A2*) = Type<A3>::scan;
  if (f3) f3(&a3);
  void(*const f4)(A4*) = Type<A4>::scan;
  if (f4) f4(&a4);
}

template <class R, class A1, class A2, class A3, class A4>
inline void 
ArgBagThread4<R, A1, A2, A3, A4>::
init(ThreadBag<R> _bag, 
     R (*_f)(A1, A2, A3, A4), A1 _a1, A2 _a2, A3 _a3, A4 _a4)
{
  ResBagThread<R>::init();
  f = _f;
  a1 = _a1;
  a2 = _a2;
  a3 = _a3;
  a4 = _a4;
  userfun = (void(*)(CoreThread*))user;
  scanfun = (void(*)(CoreThread*))ArgBagThread4<R, A1, A2, A3, A4>::scan;
  add(_bag);
}

// ----------------------------------------------------------------------------
//
// bag thread pointer with no argument
//
// ----------------------------------------------------------------------------
template <class R>
inline ArgBagThread0<R> *
ThreadBag0<R>::init(ThreadBag<R> bag, R (*f)(void))
{
  ArgBagThread0<R> *t = 
    (ArgBagThread0<R>*)
      processor.alloc(sizeof(ArgBagThread0<R>));
  t->init(bag, f);
  return(t);
}

template <class R>
inline void
ThreadBag0<R>::create(ThreadBag<R> bag, R (*f)(void))
{
  ArgBagThread0<R> *t = init(bag, f);
#ifndef RT_TBAG_NOTPAR
  sys->is_lazy(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R>
inline void
ThreadBag0<R>::start(ThreadBag<R> bag, R (*f)(void))
{
  ArgBagThread0<R> *t = init(bag, f);
#ifndef RT_TBAG_NOTPAR
  t->realize();
  sys->is_ready(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R>
inline
ThreadBag0<R>::ThreadBag0(ThreadBag<R> bag, R (*f)(void))
{
#ifdef RT_TBAG_NOTLAZY 
  start(bag, f);   // real thread
#else		       
  create(bag, f);  // lazy thread
#endif
}

// ----------------------------------------------------------------------------
//
// bag thread pointer with one argument
//
// ----------------------------------------------------------------------------
template <class R, class A1>
inline ArgBagThread1<R, A1> *
ThreadBag1<R, A1>::init(ThreadBag<R> bag, R (*f)(A1), A1 a1)
{
  ArgBagThread1<R, A1> *t = 
    (ArgBagThread1<R, A1>*)
      processor.alloc(sizeof(ArgBagThread1<R, A1>));
  t->init(bag, f, a1);
  return(t);
}

template <class R, class A1>
inline void
ThreadBag1<R, A1>::create(ThreadBag<R> bag, R (*f)(A1), A1 a1)
{
  ArgBagThread1<R, A1> *t = init(bag, f, a1);
#ifndef RT_TBAG_NOTPAR
  sys->is_lazy(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1>
inline void
ThreadBag1<R, A1>::start(ThreadBag<R> bag, R (*f)(A1), A1 a1)
{
  ArgBagThread1<R, A1> *t = init(bag, f, a1);
#ifndef RT_TBAG_NOTPAR
  t->realize();
  sys->is_ready(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1>
inline
ThreadBag1<R, A1>::ThreadBag1(ThreadBag<R> bag, R (*f)(A1), A1 a1)
{
#ifdef RT_TBAG_NOTLAZY 
  start(bag, f, a1);   // real thread
#else		       
  create(bag, f, a1);  // lazy thread
#endif
}

// ----------------------------------------------------------------------------
//
// bag thread pointer with two arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2>
inline ArgBagThread2<R, A1, A2> *
ThreadBag2<R, A1, A2>::
init(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2)
{
  ArgBagThread2<R, A1, A2> *t = 
    (ArgBagThread2<R, A1, A2>*)
      processor.alloc(sizeof(ArgBagThread2<R, A1, A2>));
  t->init(bag, f, a1, a2);
  return(t);
}

template <class R, class A1, class A2>
inline void
ThreadBag2<R, A1, A2>::
create(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2)
{
  ArgBagThread2<R, A1, A2> *t = init(bag, f, a1, a2);
#ifndef RT_TBAG_NOTPAR
  sys->is_lazy(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2>
inline void
ThreadBag2<R, A1, A2>::
start(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2)
{
  ArgBagThread2<R, A1, A2> *t = init(bag, f, a1, a2);
#ifndef RT_TBAG_NOTPAR
  t->realize();
  sys->is_ready(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2>
inline
ThreadBag2<R, A1, A2>::
ThreadBag2(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2)
{
#ifdef RT_TBAG_NOTLAZY 
  start(bag, f, a1, a2);   // real thread
#else		       
  create(bag, f, a1, a2);  // lazy thread
#endif
}

// ----------------------------------------------------------------------------
//
// bag thread pointer with three arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2, class A3>
inline ArgBagThread3<R, A1, A2, A3> *
ThreadBag3<R, A1, A2, A3>::
init(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3)
{
  ArgBagThread3<R, A1, A2, A3> *t = 
    (ArgBagThread3<R, A1, A2, A3>*)
      processor.alloc(sizeof(ArgBagThread3<R, A1, A2, A3>));
  t->init(bag, f, a1, a2, a3);
  return(t);
}

template <class R, class A1, class A2, class A3>
inline void
ThreadBag3<R, A1, A2, A3>::
create(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3)
{
  ArgBagThread3<R, A1, A2, A3> *t = init(bag, f, a1, a2, a3);
#ifndef RT_TBAG_NOTPAR
  sys->is_lazy(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2, class A3>
inline void
ThreadBag3<R, A1, A2, A3>::
start(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3)
{
  ArgBagThread3<R, A1, A2, A3> *t = init(bag, f, a1, a2, a3);
#ifndef RT_TBAG_NOTPAR
  t->realize();
  sys->is_ready(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2, class A3>
inline
ThreadBag3<R, A1, A2, A3>::
ThreadBag3(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3)
{
#ifdef RT_TBAG_NOTLAZY 
  start(bag, f, a1, a2, a3);   // real thread
#else		       
  create(bag, f, a1, a2, a3);  // lazy thread
#endif
}

// ----------------------------------------------------------------------------
//
// bag thread pointer with four arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2, class A3, class A4>
inline ArgBagThread4<R, A1, A2, A3, A4> *
ThreadBag4<R, A1, A2, A3, A4>::
init(ThreadBag<R> bag, 
     R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4)
{
  ArgBagThread4<R, A1, A2, A3, A4> *t = 
    (ArgBagThread4<R, A1, A2, A3, A4>*)
      processor.alloc(sizeof(ArgBagThread4<R, A1, A2, A3, A4>));
  t->init(bag, f, a1, a2, a3, a4);
  return(t);
}

template <class R, class A1, class A2, class A3, class A4>
inline void
ThreadBag4<R, A1, A2, A3, A4>::
create(ThreadBag<R> bag, 
       R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4)
{
  ArgBagThread4<R, A1, A2, A3, A4> *t = init(bag, f, a1, a2, a3, a4);
#ifndef RT_TBAG_NOTPAR
  sys->is_lazy(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2, class A3, class A4>
inline void
ThreadBag4<R, A1, A2, A3, A4>::
start(ThreadBag<R> bag, 
       R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4)
{
  ArgBagThread4<R, A1, A2, A3, A4> *t = init(bag, f, a1, a2, a3, a4);
#ifndef RT_TBAG_NOTPAR
  t->realize();
  sys->is_ready(t);
#endif
  processor.unlock();
#ifdef RT_TBAG_NOTPAR
  t->user();
#endif
}

template <class R, class A1, class A2, class A3, class A4>
inline
ThreadBag4<R, A1, A2, A3, A4>::
ThreadBag4(ThreadBag<R> bag, 
	   R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4)
{
#ifdef RT_TBAG_NOTLAZY 
  start(bag, f, a1, a2, a3, a4);   // real thread
#else		       
  create(bag, f, a1, a2, a3, a4);  // lazy thread
#endif
}

#endif
// ----------------------------------------------------------------------------
// end of rt_tbag.i
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
