// ----------------------------------------------------------------------------
//
// rt_machine.cc
// RT++ machine-specific interface
//
// $Id: rt_machine.cc,v 1.3 1996/03/22 08:46:17 ws Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

// ----------------------------------------------------------------------------
//
// 1. process forking
//    working: fork()
//    unused: Convex SPP cnx_sc_fork()
// 2. shared memory 
//    working: IPC everywhere, mmap on Sequent 
//    unused: Convex SPP IPC and mmap with detaching shm before forking
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//
// 1. process forking
//
// ----------------------------------------------------------------------------

#if 1
// ----------------------------------------------------------------------------
//
// generic: this is the default
//
// ----------------------------------------------------------------------------

void rt_fork_start(void)
{
}

pid_t rt_fork(void)
{
  return(fork());
}

void rt_fork_end(void)
{
}
#endif

#if 0
// ----------------------------------------------------------------------------
//
// Convex SPP: fork processes across subcomplexes
// - a subcomplex is a cluster of processes
//   (subcomplex ids can be seen by the Convex command <scm -c>)
// - <rt_sc_add> iteratively declares subcomplexes before <rt_main> is called
//   (if no subcomplex is declared, processes are forked in current subcomplex)
// - <rt_sc_reset> resets subcomplex declarations
// UNFORTUNATELY the SPP does not provide coherent shared memory
// among processes on different nodes i.e. the following is not necessary
//
// ----------------------------------------------------------------------------
#ifdef RT_CONVEX

#define MAX_SC 8

// declaration of subcomplexes
static int sc_nr = 0;             // number of declared subcomplexes
static cnx_scid_t sc_id[MAX_SC];  // identifiers of subcomplexes
static int sc_proc[MAX_SC];       // processor number of each subcomplex

// static information used during iteration of rt_fork calls
static int sc;   // id of current subcomplex
static int proc; // number of forked processes in current subcomplex

// ----------------------------------------------------------------------------
//
// called by RT++ before processes are forked
//
// ----------------------------------------------------------------------------
void rt_fork_start(void)
{
  sc = 0;
  proc = 0;
}

// ----------------------------------------------------------------------------
//
// called by RT++ to fork a new process
//
// ----------------------------------------------------------------------------
pid_t rt_fork(void)
{
  if (sc_nr == 0) return(fork());
  while (proc == sc_proc[sc])
    {
      proc = 0;
      sc++;
      if (sc == sc_nr) sc = 0;
    }
  proc++;
  return(cnx_sc_fork(sc_id[sc], CNX_ANY_NODE));
}

// ----------------------------------------------------------------------------
//
// called by RT++ after processes have been forked
//
// ----------------------------------------------------------------------------
void rt_fork_end(void)
{
}

// ----------------------------------------------------------------------------
//
// add a processor subcomplex with identifier <id> and <proc> processors
//
// ----------------------------------------------------------------------------
void rt_sc_add(cnx_scid_t id, int proc)
{
  if (sc_nr == MAX_SC) rt_panic("too many subcomplexes");
  if (proc <= 0) rt_panic("subcomplex processor number not greater than 0");
  sc_id[sc_nr] = id;
  sc_proc[sc_nr] = proc;
  sc_nr++;
}

// ----------------------------------------------------------------------------
//
// reset subcomplex declaration
//
// ----------------------------------------------------------------------------
void rt_sc_reset(void)
{
  sc_nr = 0;
}
#endif
#endif

// ----------------------------------------------------------------------------
//
// 2. shared memory
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//
// SysV IPC
//
// ----------------------------------------------------------------------------
#ifdef RT_IPC

#if 1
// ----------------------------------------------------------------------------
//
// IPC generic version: this is the default shared memory mechanism
//
// ----------------------------------------------------------------------------

static int shm_id;     // IPC identifier
static void *shm_addr; // IPC shared memory address 

void *rt_shm_alloc(int n)
{
  // allocate shared memory segment
  shm_id = shmget(IPC_PRIVATE, n, 0600);
  if (shm_id == -1) rt_panic("shared memory allocation failed");
  // attach shared memory segment into the space of the current process
  shm_addr = shmat(shm_id, 0, 0);
  if (shm_addr == (void*)-1) rt_panic("shared memory mapping failed");
  return(shm_addr);
}

void rt_shm_unmap(void)
{

}

void rt_shm_map(void)
{

}

void rt_shm_close(void)
{
  // remove shared memory segment
  shmctl(shm_id, IPC_RMID, 0);
}

void rt_shm_free(void)
{
}
#endif

#if 0
// ----------------------------------------------------------------------------
//
// IPC on Convex SPP: unmap shared memory before forking and map again
// UNFORTUNATELY SPP does not provide coherent shared memory among processes
// on different nodes i.e. this is superfluous
//
// ----------------------------------------------------------------------------
#ifdef RT_CONVEX

static int shm_id;
static caddr_t shm_addr;
static int shm_len;

void *rt_shm_alloc(int n)
{
  shm_id = shmget(IPC_PRIVATE, n, 0600);
  if (shm_id == -1) rt_panic("shared memory allocation failed");
  shm_addr = shmat(id, 0, 0);
  if (shm_addr == (caddr_t)-1) rt_panic("shared memory mapping failed");
  shm_len = n;
  return(s);
}

void rt_shm_unmap(void)
{
  int code = shmdt(shm_addr);
  if (code == -1) rt_panic("shared memory detaching failed");
}

void rt_shm_map(void)
{
  void *s = shmat(shm_id, shm_addr, 0);
  if (s == (void*)-1) rt_panic("shared memory attaching failed");
}

void rt_shm_close(void)
{
}

void rt_shm_free(void)
{  
  shmctl(shm_id, IPC_RMID, 0);
}
#endif
#endif

#endif

// ----------------------------------------------------------------------------
//
// BSD mmap
//
// ----------------------------------------------------------------------------
#if defined(RT_MMAP)

// ----------------------------------------------------------------------------
//
// Sequent Dynix/ptx: consider alignment requirements
//
// ----------------------------------------------------------------------------
#if defined(RT_PTX)

static caddr_t shm_addr;
static int shm_len;

void *rt_shm_alloc(int n)
{
  int align = sysconf(_SC_PAGE_SIZE)-1;
  shm_len = (n+align) & ~align;
  shm_addr = mmap(0, shm_len, PROT_READ | PROT_WRITE, 
		  MAP_ANONYMOUS | MAP_VARIABLE | MAP_SHARED, 
		  0, 0);
  if (shm_addr == (caddr_t)-1) rt_panic("shared memory file mapping failed");
  return(shm_addr);
}

void rt_shm_unmap(void)
{
}

void rt_shm_map(void)
{
  
}

void rt_shm_close(void)
{

}

void rt_shm_free(void)
{
  int code = munmap(shm_addr, shm_len);
  if (code == -1) rt_panic("memory mapped file could not be unmapped");
}
#endif

#if 0
// ----------------------------------------------------------------------------
//
// mmap on Convex SPP: unmap shared memory before forking and map again
// UNFORTUNATELY SPP does not provide coherent shared memory among processes
// on different nodes i.e. this idea does not work
//
// ----------------------------------------------------------------------------
#ifdef RT_CONVEX

static caddr_t shm_addr;
static int shm_len;
static char shm_fd;
static char shm_file[L_tmpnam];

void *rt_shm_alloc(int n)
{
  tmpnam(shm_file);
  shm_fd = creat(shm_file, 0666);
  if (shm_fd == -1) rt_panic("shared memory file could not be created");
  int m = lseek(shm_fd, n, SEEK_SET);
  if (m != n) rt_panic("shared memory file pointer could not be set");
  int wc = write(shm_fd, "end\n", 4);
  if (wc == -1) rt_panic("shared memory file could not be written");
  int lc = close(shm_fd);
  if (lc == -1) rt_panic("shared memory file could not be closed");
  shm_fd = open(shm_file, O_RDWR);
  if (shm_fd == -1) rt_panic("shared memory file could not be opened");
  int uc = unlink(shm_file);
  if (uc == -1) rt_panic("shared memory file could not be unlinked");
  shm_addr = mmap(0, n, PROT_READ | PROT_WRITE, 
		  MAP_FILE | MAP_VARIABLE | MAP_SHARED, shm_fd, 0);
  if (shm_addr == (caddr_t)-1) rt_panic("shared memory file mapping failed");
  shm_len = n;
  return(s);
}

void rt_shm_unmap(void)
{
  int uc = munmap(shm_addr, shm_len);
  if (uc == -1) rt_panic("shared memory region could not be unmapped");
}

void rt_shm_map(void)
{
  void *s = mmap(shm_addr, shm_len, PROT_READ | PROT_WRITE, 
		 MAP_FILE | MAP_FIXED | MAP_SHARED,
		 shm_fd, 0);
  if (s == (void*)-1) rt_panic("shared memory mapping not identical");
}

void rt_shm_close(void)
{
  close(shm_fd);
}

void rt_shm_free(void)
{  
  int code = munmap(shm_addr, shm_len);
  if (code == -1) rt_panic("memory mapped file could not be unmapped");
}
#endif
#endif

#endif

// ----------------------------------------------------------------------------
// end of rt_machine.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
